import enum
import logging
import math
from PyQt5 import QtCore
from PyQt5 import QtWidgets
from PyQt5 import QtGui
import matc.constants
import matc.globa
import matc.settings

WINDOW_FLAGS = (
    QtCore.Qt.Dialog
    | QtCore.Qt.FramelessWindowHint
    | QtCore.Qt.WindowStaysOnTopHint
)
# With this it seems we don't capture keyboard events: QtCore.Qt.WindowDoesNotAcceptFocus
# QtCore.Qt.BypassWindowManagerHint


class CursorPosition(enum.Enum):
    inner = enum.auto()
    outside = enum.auto()


class BreathingState(enum.Enum):
    inactive = 0
    breathing_in = 1
    breathing_out = 2


class BreathingDlg(QtWidgets.QDialog):
    close_signal = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.setWindowFlags(WINDOW_FLAGS)
        self.setWindowTitle(f"Breathing Dialog - {matc.constants.APPLICATION_PRETTY_NAME}")
        self.setWindowIcon(QtGui.QIcon(matc.globa.get_app_icon_path("icon.png")))
        self.setStyleSheet(
            f"background-color: {matc.globa.BLACK_COLOR};"
            "color: #999999;"
            f"selection-background-color: {matc.globa.LIGHT_GREEN_COLOR};"
            "selection-color:#000000;"
        )

        self.vbox_l2 = QtWidgets.QVBoxLayout()
        self.vbox_l2.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.vbox_l2)

        self.breathing_gv = BreathingGraphicsView()
        self.vbox_l2.addWidget(self.breathing_gv, alignment=QtCore.Qt.AlignHCenter)
        self.breathing_gv.mouse_btn_released_signal.connect(self.on_brdlg_mouse_btn_released)
        self.breathing_gv.mouse_leave_signal.connect(self.on_brdlg_mouse_leave)
        self.breathing_gv.return_pressed_signal.connect(self.on_brdlg_return_pressed)
        # self.breathing_gv.initiate()

    def close_dlg(self):
        # self.showNormal()
        # -for MacOS. showNormal is used here rather than showMinimized to avoid animation
        self.breathing_gv.ib_qtimeline.stop()
        self.breathing_gv.ob_qtimeline.stop()
        # self.breathing_dlg.circle_go.setScale(1)
        self.breathing_gv.hide()
        self.hide()
        self.close_signal.emit()

    def on_brdlg_mouse_btn_released(self):
        self.close_dlg()

    def on_brdlg_mouse_leave(self):
        self.close_dlg()

    def on_brdlg_return_pressed(self):
        self.close_dlg()

    def show_breathing_only(self):
        # self.breathing_dlg.breathing_gi.update_pos_and_origin_point(VIEW_WIDTH_INT, VIEW_HEIGHT_INT)

        if self.isVisible():
            return

        self.adjustSize()  # -also seems to set x and y to 0
        screen_qrect = QtWidgets.QApplication.desktop().availableGeometry()
        _xpos_int = screen_qrect.left() + (screen_qrect.width() - VIEW_WIDTH_INT) // 2
        _ypos_int = screen_qrect.bottom() - VIEW_HEIGHT_INT + 1
        # -self.sizeHint().height() gives only 52 here, unknown why, so we use VIEW_HEIGHT_INT instead
        # logging.debug("screen_qrect.bottom() = " + str(screen_qrect.bottom()))
        # logging.debug("self.sizeHint().height() = " + str(self.sizeHint().height()))
        self.move(_xpos_int, _ypos_int)
        # self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        # self.breathing_dlg.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self.close_dlg()
        self.showNormal()

        self.breathing_gv.initiate()

    def on_close_clicked(self):
        self.close_dlg()


TIME_NOT_SET_FT = 0.0
TIME_LINE_IB_DURATION_INT = 8000
TIME_LINE_OB_DURATION_INT = 16000
TIME_LINE_DOT_DURATION_INT = 1000
TIME_LINE_IB_FRAME_RANGE_INT = 1000
TIME_LINE_OB_FRAME_RANGE_INT = 2000
TIME_LINE_IB_DOT_FRAME_RANGE_INT = 255

VIEW_WIDTH_INT = 540
VIEW_HEIGHT_INT = 270
BR_CIRCLE_RADIUS_FT = 45.0

DOT_RADIUS_FT = 7
DOT_SPACING = 3


class BreathingGraphicsView(QtWidgets.QGraphicsView):
    mouse_btn_released_signal = QtCore.pyqtSignal()
    mouse_leave_signal = QtCore.pyqtSignal()
    first_breathing_gi_signal = QtCore.pyqtSignal()
    return_pressed_signal = QtCore.pyqtSignal()

    # Also contains the graphics scene
    def __init__(self, i_can_be_closed: bool = True) -> None:
        super().__init__()

        self.breathing_count = 0
        self.br_dots_gi_list = []

        self.first_shown: bool = True
        self.breathing_state = BreathingState.inactive
        self._can_be_closed_bool = i_can_be_closed
        self._keyboard_active_bool = True
        self._start_time_ft = TIME_NOT_SET_FT
        self._ib_length_ft_list = []
        self._ob_length_ft_list = []

        # Window setup
        self.setFrameStyle(QtWidgets.QFrame.Box | QtWidgets.QFrame.Plain)
        self.setLineWidth(0)
        # self.setStyleSheet("border-radius: 15px;")
        self.setFixedWidth(VIEW_WIDTH_INT)
        self.setFixedHeight(VIEW_HEIGHT_INT)
        t_brush = QtGui.QBrush(QtGui.QColor(matc.globa.BLACK_COLOR))
        self.setBackgroundBrush(t_brush)
        self.setRenderHints(
            QtGui.QPainter.Antialiasing |
            QtGui.QPainter.SmoothPixmapTransform
        )
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        # ..set position
        screen_qrect = QtWidgets.QApplication.desktop().availableGeometry()
        self._xpos_int = screen_qrect.left() + (screen_qrect.width() - VIEW_WIDTH_INT) // 2
        self._ypos_int = screen_qrect.bottom() - VIEW_HEIGHT_INT - 60
        # -self.sizeHint().height() gives only 52 here, unknown why, so we use VIEW_HEIGHT_INT instead
        self.move(self._xpos_int, self._ypos_int)

        # Graphics and layout setup..
        vbox_l2 = QtWidgets.QVBoxLayout()
        self.setLayout(vbox_l2)
        # ..graphics scene
        self._graphics_scene = QtWidgets.QGraphicsScene()
        self._graphics_scene.setSceneRect(QtCore.QRectF(0, 0, VIEW_WIDTH_INT, VIEW_HEIGHT_INT))
        self.setScene(self._graphics_scene)

        # ..help text
        help_text_str = "Hover over the green box breathing in and outside the green box breathing out"
        self.help_text_gi = MyQGraphicsTextItem()
        self.help_text_gi.setHtml(matc.globa.get_html(help_text_str, 18))
        self._graphics_scene.addItem(self.help_text_gi)

        # ..text
        self.text_gi = MyQGraphicsTextItem()
        # self.text_gi.setAcceptHoverEvents(False)  # -so that the underlying item will not be disturbed
        self.text_gi.position_signal.connect(self._breathing_gi_position_changed)
        self._graphics_scene.addItem(self.text_gi)

        # Animation
        self.ib_qtimeline = QtCore.QTimeLine(duration=TIME_LINE_IB_DURATION_INT)
        self.ib_qtimeline.setFrameRange(1, TIME_LINE_IB_FRAME_RANGE_INT)
        self.ib_qtimeline.setCurveShape(QtCore.QTimeLine.LinearCurve)
        self.ib_qtimeline.frameChanged.connect(self.frame_change_breathing_in)
        self.ob_qtimeline = QtCore.QTimeLine(duration=TIME_LINE_OB_DURATION_INT)
        self.ob_qtimeline.setFrameRange(1, TIME_LINE_OB_FRAME_RANGE_INT)
        self.ob_qtimeline.setCurveShape(QtCore.QTimeLine.LinearCurve)
        self.ob_qtimeline.frameChanged.connect(self.frame_change_breathing_out)

        self.breathing_phrase = None
        ##### self.initiate()

        # https://forum.qt.io/topic/106003/how-to-seamlessly-place-item-into-scene-at-specific-location-adding-qgraphicsitem-to-scene-always-places-it-at-0-0/2


        # ..custom dynamic breathing graphic (may be possible to change this in the future)

        self.br_vis_go_list = []

        bar_go = BreathingBarQgo()
        bar_go.hide()
        self.br_vis_go_list.append(bar_go)
        self._graphics_scene.addItem(bar_go)

        c_go = BreathingCircleQgo()
        c_go.hide()
        self.br_vis_go_list.append(c_go)
        self._graphics_scene.addItem(c_go)

        self.dot_qtimeline = QtCore.QTimeLine(duration=TIME_LINE_DOT_DURATION_INT)
        self.dot_qtimeline.setFrameRange(1, TIME_LINE_IB_DOT_FRAME_RANGE_INT)
        # self.dot_qtimeline.setCurveShape(QtCore.QTimeLine.EaseOutCurve)
        # self.dot_qtimeline.setEasingCurve(QtCore.QEasingCurve.OutQuad)
        self.dot_qtimeline.setEasingCurve(QtCore.QEasingCurve.InOutQuad)
        self.dot_qtimeline.frameChanged.connect(self.on_frame_change_dot)

    def on_text_hover(self):
        self._breathing_gi_position_changed(CursorPosition.inner.value)

    def show(self):
        raise Exception("Call not supported, call function in super class instead, or use initiate")

    def initiate(self):
        self.breathing_state = BreathingState.inactive

        self.breathing_count = 0
        for br_dot in self.br_dots_gi_list:
            self._graphics_scene.removeItem(br_dot)
        self.br_dots_gi_list.clear()

        for br_vis_go in self.br_vis_go_list:
            br_vis_go.hide()
        br_vis: int = matc.settings.settings[matc.settings.SK_BREATHING_VISUALIZATION]
        self.active_br_go = self.br_vis_go_list[br_vis]
        self.active_br_go.show()
        self.active_br_go.update_pos_and_origin_point()
        self.active_br_go.position_signal.connect(self._breathing_gi_position_changed)

        if matc.globa.active_phrase_id == matc.globa.BREATHING_PHRASE_NOT_SET:
            self.breathing_phrase = matc.settings.get_topmost_breathing_phrase()
            matc.globa.active_phrase_id = self.breathing_phrase.id
        self.breathing_phrase = matc.settings.get_breathing_phrase(matc.globa.active_phrase_id)

        help_text_pointf = QtCore.QPointF(
            VIEW_WIDTH_INT / 2 - self.help_text_gi.boundingRect().width() / 2,
            10
        )
        self.help_text_gi.setPos(help_text_pointf)
        self.help_text_gi.show()

        self.text_gi.setHtml(self._get_ib_ob_html())
        text_pointf = QtCore.QPointF(
            VIEW_WIDTH_INT / 2 - self.text_gi.boundingRect().width() / 2,
            VIEW_HEIGHT_INT - self.text_gi.boundingRect().height() - 10
        )
        self.text_gi.setPos(text_pointf)

        super().show()

        move_mouse_cursor: bool = matc.settings.settings[matc.settings.SK_MOVE_MOUSE_CURSOR]
        if self.first_shown:
            self.first_shown = False
        elif move_mouse_cursor:
            screen_point = self.mapToGlobal(QtCore.QPoint(35, 35))
            screen = QtGui.QGuiApplication.primaryScreen()
            mouse_cursor = QtGui.QCursor()
            mouse_cursor.setPos(screen, screen_point)
            # https://doc.qt.io/qt-5/qcursor.html#setPos-1

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mousePressEvent(event)
        event.accept()
        if self._can_be_closed_bool:
            self.mouse_btn_released_signal.emit()

    def leaveEvent(self, i_qevent) -> None:
        if self._can_be_closed_bool:
            self.mouse_leave_signal.emit()

    def _get_ib_ob_html(self, i_ib_underscore: bool = False, i_ob_underscore: bool = False) -> str:
        ib_html: str = matc.globa.get_html(self.breathing_phrase.in_breath, 18, i_ib_underscore)
        ob_html: str = matc.globa.get_html(self.breathing_phrase.out_breath, 18, i_ob_underscore)
        return ib_html + ob_html

    def _start_breathing_in(self) -> None:
        if self.breathing_state == BreathingState.breathing_in:
            return
        self.breathing_state = BreathingState.breathing_in

        self.on_frame_change_dot(TIME_LINE_IB_DOT_FRAME_RANGE_INT)

        self.breathing_count += 1
        br_dots_gi = DotQgo(self.breathing_count)
        self.br_dots_gi_list.append(br_dots_gi)
        self._graphics_scene.addItem(br_dots_gi)
        br_dots_gi.show()
        for br_dot in self.br_dots_gi_list:
            br_dot.update_pos(len(self.br_dots_gi_list))

        if self.breathing_count == 1:
            self.first_breathing_gi_signal.emit()

        self.help_text_gi.hide()

        self.text_gi.setHtml(self._get_ib_ob_html(i_ib_underscore=True))
        self.ob_qtimeline.stop()
        self.ib_qtimeline.start()

        self.dot_qtimeline.stop()
        self.dot_qtimeline.start()

    def _start_breathing_out(self) -> None:
        if self.breathing_state != BreathingState.breathing_in:
            return
        self.breathing_state = BreathingState.breathing_out

        self.active_br_go._peak_width_ft = self.active_br_go.rectf.width()

        self.text_gi.setHtml(self._get_ib_ob_html(i_ob_underscore=True))
        self.ib_qtimeline.stop()
        self.ob_qtimeline.start()

    def keyPressEvent(self, i_qkeyevent) -> None:
        if self._keyboard_active_bool:
            if i_qkeyevent.key() == QtCore.Qt.Key_Shift:
                logging.debug("shift key pressed")
                self._start_breathing_in()
            elif i_qkeyevent.key() == QtCore.Qt.Key_Return:
                logging.debug("return key pressed")
                self.return_pressed_signal.emit()

    def keyReleaseEvent(self, i_qkeyevent) -> None:
        if self._keyboard_active_bool:
            if i_qkeyevent.key() == QtCore.Qt.Key_Shift:
                logging.debug("shift key released")
                self._start_breathing_out()

    def _breathing_gi_position_changed(self, i_pos_type: int) -> None:
        if i_pos_type == CursorPosition.inner.value:
            self._start_breathing_in()
            # self.circle_go.update_pos_and_origin_point(VIEW_WIDTH_INT, VIEW_HEIGHT_INT)

        elif i_pos_type == CursorPosition.outside.value:
            self._start_breathing_out()
            # self.text_gi.update_pos_and_origin_point(VIEW_WIDTH_INT, VIEW_HEIGHT_INT)
            # self.circle_go.update_pos_and_origin_point(VIEW_WIDTH_INT, VIEW_HEIGHT_INT)

        """
        # Calculating coords..
        # ..breathing rectangle/area
        hover_rectangle_qsize = QtCore.QSizeF(BR_WIDTH_FT, BR_HEIGHT_FT)
        x: float = self.circle_go.x() + (self.circle_go.boundingRect().width() - hover_rectangle_qsize.width()) / 2
        y: float = self.circle_go.y() + (self.circle_go.boundingRect().height() - hover_rectangle_qsize.height()) / 2
        pos_pointf = QtWidgets.QGraphicsItem.mapFromItem(self.circle_go, self.circle_go, x, y)  # -widget coords
        hover_rectangle_coords_qrect = QtCore.QRectF(pos_pointf, hover_rectangle_qsize)
        logging.debug(f"{hover_rectangle_coords_qrect=}")
        # ..mouse cursor
        cursor = QtGui.QCursor()  # -screen coords
        cursor_pos_widget_coords_qp = self.mapFromGlobal(cursor.pos())  # -widget coords
        logging.debug(f"{cursor_pos_widget_coords_qp=}")

        # Check if cursor is inside the area, and start the animation
        if hover_rectangle_coords_qrect.contains(cursor_pos_widget_coords_qp):
            matc.globa.breathing_state = matc.globa.BreathingState.breathing_in
            self._start_breathing_in()
            self.circle_go.update_pos_and_origin_point(VIEW_WIDTH_INT, VIEW_HEIGHT_INT)
        """

    def frame_change_breathing_in(self, i_frame_nr: int) -> None:
        # new_scale_int_ft = 1 + 0.001 * i_frame_nr
        # self.circle_go.setScale(new_scale_int_ft)

        self.active_br_go.change_size_br_in(i_frame_nr)

    def frame_change_breathing_out(self, i_frame_nr: int) -> None:
        # self.circle_go.setScale(new_scale_int_ft)

        self.active_br_go.change_size_br_out(i_frame_nr)

    def on_frame_change_dot(self, i_frame_nr: int) -> None:
        if len(self.br_dots_gi_list) <= 0:
            return
        last_dot_git: DotQgo = self.br_dots_gi_list[-1]
        last_dot_git.color.setAlpha(i_frame_nr)
        last_dot_git.update()
        # self._graphics_scene.update(last_dot_git.rectf)
        # last_dot_git.paint()


class MyQGraphicsTextItem(QtWidgets.QGraphicsTextItem):
    position_signal = QtCore.pyqtSignal(int)

    def __init__(self) -> None:
        super().__init__()
        self.setDefaultTextColor(QtGui.QColor(matc.globa.DARKER_GREEN_COLOR))
        self.setTextWidth(VIEW_WIDTH_INT - 20)

    def hoverMoveEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        cursor_posx = int(event.pos().x())
        cursor_posy = int(event.pos().y())
        upper_rect = QtCore.QRect(
            int(self.boundingRect().x()), int(self.boundingRect().y()),
            int(self.boundingRect().width()), int(self.boundingRect().height() // 2))
        if upper_rect.contains(cursor_posx, cursor_posy):
            self.position_signal.emit(CursorPosition.inner.value)
        else:
            self.position_signal.emit(CursorPosition.outside.value)

    def hoverLeaveEvent(self, i_qgraphicsscenehoverevent) -> None:
        # Please note that this function is entered in case the user hovers over something
        #  on top of this graphics item
        self.position_signal.emit(CursorPosition.outside.value)


class DotQgo(QtWidgets.QGraphicsObject):
    def __init__(self, i_number: int):
        super().__init__()
        self.number = i_number
        self.rectf = QtCore.QRectF(0, 0, 2*DOT_RADIUS_FT, 2*DOT_RADIUS_FT)
        self.setAcceptHoverEvents(False)
        self.color = QtGui.QColor(matc.globa.LIGHT_GREEN_COLOR)
        self.color.setAlpha(0)

    def boundingRect(self):
        return self.rectf

    # Overridden
    def paint(self, i_qpainter: QtGui.QPainter, i_qstyleoptiongraphicsitem, widget=None) -> None:
        # i_qpainter.fillRect(self.rectf, t_brush)
        t_brush = QtGui.QBrush(self.color)
        i_qpainter.setBrush(t_brush)
        pen = QtGui.QPen()
        pen.setWidth(0)
        i_qpainter.setPen(pen)
        i_qpainter.drawEllipse(self.rectf)

    def update_pos(self, i_total_nr: int) -> None:
        x_delta = (self.number - 0.5 - i_total_nr / 2) * (self.boundingRect().width() + DOT_SPACING)
        x: float = VIEW_WIDTH_INT / 2 - self.boundingRect().width() / 2 + x_delta
        y: float = self.boundingRect().height()
        self.setPos(QtCore.QPointF(x, y))


class BreathingQgo(QtWidgets.QGraphicsObject):
    position_signal = QtCore.pyqtSignal(int)

    def change_size_br_in(self, i_frame_nr: int):
        pass

    def change_size_br_out(self, i_frame_nr: int):
        pass

    def __init__(self):
        super().__init__()
        self.rectf = QtCore.QRectF()
        self.setAcceptHoverEvents(True)
        self._peak_width_ft = BR_CIRCLE_RADIUS_FT * 2

    def boundingRect(self):
        return self.rectf

    def update_pos_and_origin_point(self) -> None:
        x: float = VIEW_WIDTH_INT / 2 - self.boundingRect().width() / 2
        y: float = VIEW_HEIGHT_INT / 2 - self.boundingRect().height() / 2
        self.setPos(QtCore.QPointF(x, y))
        self.setTransformOriginPoint(self.boundingRect().center())

    def hoverLeaveEvent(self, i_qgraphicsscenehoverevent) -> None:
        # Please note that this function is entered in case the user hovers over something
        #  on top of this graphics item
        self.position_signal.emit(CursorPosition.outside.value)

    def hoverMoveEvent(self, i_qgraphicsscenehoverevent: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        pass


class BreathingCircleQgo(BreathingQgo):
    """
    "The QGraphicsObject class provides a base class for all graphics items that require signals,
    slots and properties."
    https://doc.qt.io/qt-5/qgraphicsobject.html

    breathing in: ignoring the state of the circle, instead using the same starting state
    breathing out: using the state of the circle
    """

    def __init__(self):
        super().__init__()
        self.rectf = QtCore.QRectF(
            0, 0,
            2*BR_CIRCLE_RADIUS_FT, 2*BR_CIRCLE_RADIUS_FT
        )

    def change_size_br_in(self, i_frame_nr: int):
        new_width_ft = 2 * BR_CIRCLE_RADIUS_FT + 0.1 * i_frame_nr
        self.rectf.setWidth(new_width_ft)
        self.rectf.setHeight(new_width_ft)
        self.update_pos_and_origin_point()

    def change_size_br_out(self, i_frame_nr: int):
        new_width_ft = self._peak_width_ft - 0.06 * i_frame_nr
        if new_width_ft < 2 * BR_CIRCLE_RADIUS_FT:
            new_width_ft = 2 * BR_CIRCLE_RADIUS_FT
        self.rectf.setWidth(new_width_ft)
        self.rectf.setHeight(new_width_ft)
        self.update_pos_and_origin_point()

    # Overridden
    def paint(self, i_qpainter: QtGui.QPainter, i_qstyleoptiongraphicsitem, widget=None) -> None:
        # i_qpainter.fillRect(self.rectf, t_brush)
        t_brush = QtGui.QBrush(QtGui.QColor(matc.globa.LIGHT_GREEN_COLOR))
        i_qpainter.setBrush(t_brush)
        i_qpainter.drawEllipse(self.rectf)

    # Overridden
    def hoverMoveEvent(self, i_qgraphicsscenehoverevent: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        # self.hover_signal.emit()

        cposx = self.boundingRect().center().x()
        cposy = self.boundingRect().center().y()
        # logging.debug(f"{cposy=}")
        pposx = i_qgraphicsscenehoverevent.pos().x()
        pposy = i_qgraphicsscenehoverevent.pos().y()
        # logging.debug(f"{pposy=}")

        distance_from_center: float = math.dist([0, 0], [pposx-cposx, pposy-cposy])
        # logging.debug(f"{distance_from_center=}")

        if distance_from_center < BR_CIRCLE_RADIUS_FT:
            self.position_signal.emit(CursorPosition.inner.value)
        elif distance_from_center > self.rectf.width() // 2:
            self.position_signal.emit(CursorPosition.outside.value)


BR_BAR_WIDTH_FT = 120.0
BR_BAR_HEIGHT_FT = 40.0


class BreathingBarQgo(BreathingQgo):
    def __init__(self):
        super().__init__()
        self.rectf = QtCore.QRectF(
            0, 0,
            BR_BAR_WIDTH_FT, BR_BAR_HEIGHT_FT
        )

    # Overridden
    def paint(self, i_qpainter: QtGui.QPainter, i_qstyleoptiongraphicsitem, widget=None) -> None:
        # i_qpainter.fillRect(self.rectf, t_brush)
        t_brush = QtGui.QBrush(QtGui.QColor(matc.globa.LIGHT_GREEN_COLOR))
        i_qpainter.setBrush(t_brush)
        i_qpainter.drawRoundedRect(self.rectf, 5, 5)

    def change_size_br_in(self, i_frame_nr: int):
        new_width_ft = BR_BAR_WIDTH_FT + 0.2 * i_frame_nr
        self.rectf.setWidth(new_width_ft)
        self.update_pos_and_origin_point()

    def change_size_br_out(self, i_frame_nr: int):
        new_width_ft = self._peak_width_ft - 0.12 * i_frame_nr
        if new_width_ft < BR_BAR_WIDTH_FT:
            new_width_ft = BR_BAR_WIDTH_FT
        self.rectf.setWidth(new_width_ft)
        self.update_pos_and_origin_point()

    # Overridden
    def hoverMoveEvent(self, i_qgraphicsscenehoverevent: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        self.position_signal.emit(CursorPosition.inner.value)



import logging
from PyQt5 import QtCore
from PyQt5 import QtWidgets
from PyQt5 import QtGui
import matc.gui.breathing_dlg
import matc.globa
import matc.settings
import matc.constants

NEXT = "Next >>"
PREV = "<< Prev"
MARGIN_TOP = 35
WIDGET_SPACING = 10

"""
An alternative to using a custom QDialog can be to use QWizard with QWizardPages
"""


class Label(QtWidgets.QLabel):
    def __init__(self, i_text: str, i_font_size: matc.globa.FontSize = matc.globa.FontSize.medium):
        super().__init__(text=i_text)
        self.setWordWrap(True)
        self.setFont(matc.globa.get_font(i_font_size))
        self.setTextFormat(QtCore.Qt.MarkdownText)
        self.setOpenExternalLinks(True)
        # text_qll.setAlignment(QtCore.Qt.AlignHCenter)


class IconImage(QtWidgets.QLabel):
    def __init__(self, i_file_name: str):
        super().__init__()
        # text_qll.setAlignment(QtCore.Qt.AlignHCenter)
        self.setPixmap(QtGui.QPixmap(matc.globa.get_app_icon_path(i_file_name)))
        # self.setAlignment(QtCore.Qt.AlignHCenter)


class IntroDlg(QtWidgets.QDialog):
    """
    The introduction wizard with examples of dialogs and functionality to adjust initial settings
    """
    close_signal = QtCore.pyqtSignal(bool)
    # -the boolean indicates whether we want the breathing dialog to open

    def __init__(self):
        super().__init__()
        self.setWindowTitle(f"Intro Dialog - {matc.constants.APPLICATION_PRETTY_NAME}")
        self.setWindowIcon(QtGui.QIcon(matc.globa.get_app_icon_path("icon.png")))

        self.wizard_qsw_w3 = QtWidgets.QStackedWidget()
        self.prev_qpb = QtWidgets.QPushButton(PREV)
        self.prev_qpb.setFont(matc.globa.get_font(matc.globa.FontSize.medium))
        self.next_qpb = QtWidgets.QPushButton(NEXT)
        self.next_qpb.setFont(matc.globa.get_font(matc.globa.FontSize.medium))

        self.prev_qpb.clicked.connect(self.on_prev_clicked)
        self.next_qpb.clicked.connect(self.on_next_clicked)

        hbox_l3 = QtWidgets.QHBoxLayout()
        hbox_l3.addStretch(1)
        hbox_l3.addWidget(self.prev_qpb, stretch=1)
        hbox_l3.addWidget(self.next_qpb, stretch=1)
        hbox_l3.addStretch(1)

        vbox_l2 = QtWidgets.QVBoxLayout()
        self.setLayout(vbox_l2)
        vbox_l2.addWidget(self.wizard_qsw_w3)
        vbox_l2.addLayout(hbox_l3)

        welcome_description_markdown_ll = Label(
            """Welcome to Mindfulness at the Computer!

This introduction will help you understand how to use the application

Mindfulness at the Computer is an application that helps you stay **mindful** while using the computer by reminding you to take **breathing** breaks. These breaks are **interactive** to help you focus on your breathing. There are also **breathing phrases** that can be used while following the breath, to stay mindful of your body, or other ascpects of your experience

The main parts of the application:
* The system tray icon
* The breathing dialog
* The settings window

These parts will now be discussed on the following pages. Please click *next* to continue"""
        )
        welcome_page = IntroPage("Welcome", welcome_description_markdown_ll)
        self.wizard_qsw_w3.addWidget(welcome_page)

        systray_description_markdown_ll = Label(
            """When you run Mindfulness at the Computer it is accessible via the system tray.
From the menu that opens when clicking on this icon you can:
* Open the settings window
* Invoke a breathing session (using the last phrase)
* Invoke a breathing session using a selected phrase
* Exit the application"""
        )

        icon_image = IconImage("icon.png")
        notification_description_markdown_ll = Label(
            f"""**The breathing notification**

This notification is shown at certain intervals to remind you to take a breathing break. You can adjust how often you would like to get this notification."""
        )

        icon_br_image = IconImage("icon-b.png")
        systray_page = IntroPage("The system tray", systray_description_markdown_ll, icon_image,
            notification_description_markdown_ll, icon_br_image)
        self.wizard_qsw_w3.addWidget(systray_page)

        br_dlg_description_ll = Label(
            """This dialog helps you to relax and return to your breathing. Try it out, it interactive!"""
        )
        self.breathing_dlg = matc.gui.breathing_dlg.BreathingGraphicsView(i_can_be_closed=False)
        self.breathing_dlg.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        self.breathing_dlg.initiate()
        br_dlg_details_ll = Label(
            """There are three ways to interact with the breathing dialog:
* Hovering over the light green area in the middle while breathing in
* Hovering over the text for the in-breath while breathing in
* Pressing down and holding shift while breathing in"""
        )
        empty_ll = Label("")  # -this is to give some extra space below the breathing widget
        self.br_dlg_page = IntroPage("The breathing dialog", br_dlg_description_ll,
            self.breathing_dlg, empty_ll, br_dlg_details_ll)
        self.wizard_qsw_w3.addWidget(self.br_dlg_page)

        settings_description_ll = Label(
            """The settings dialog can be reached by opening the systray menu and selecting "Settings" from there. (Please open it now if you want to follow along in the description below)"""
        )
        settings_details_ll = Label(
            f"""Some of the settings that can be changed:
* Amount of time before a breahitng notification is shown --- You may want to adjust this setting now (the default is {matc.settings.BREATHING_BREAK_TIMER_DEFAULT_SECS // 60} minutes)
* Volume of the audio (bells)
* Breathing phrases (displayed at the bottom of the breathing dialog) --- It is possible to add new phrases and reorder them. The topmost phrase is always used at the start
* Whether or not to move the mouse cursor into the breathing dialog (useful if you are using a touchpad)"""
        )
        settings_page = IntroPage("Settings", settings_description_ll, settings_details_ll)
        self.wizard_qsw_w3.addWidget(settings_page)

        additional_setup_ll = Label(
            """Now that you have started the application you may want to do some *additional setup*. Please find the instructions for your OS on this page: https://mindfulness-at-the-computer.gitlab.io/installation/"""
        )

        relaunch_wizard_ll = Label(
            """You can start this wizard again by choosing "Help" -> "Show intro wizard" in the settings window (available from the system tray icon menu)"""
        )
        other_help_ll = Label(
            """Other ways to get help:
* The gitter chat: https://gitter.im/mindfulness-at-the-computer/community
* Email: [sunyata.software@gmail.com](mailto:sunyata.software@gmail.com)"""
        )
        # more: feedback
        feedback_ll = Label(
            """We are grateful for any feedback you can give us. Please use the email address above to contact us with gratitudes or suggestions for improvements"""
        )
        finish_text_ll = Label(
            """When you click on finish and exit this wizard a breathing dialog will be shown."""
        )

        finish_page = IntroPage("Finish", additional_setup_ll, relaunch_wizard_ll, other_help_ll,
            feedback_ll, finish_text_ll)
        self.wizard_qsw_w3.addWidget(finish_page)

        self.setGeometry(150, 100, 680, 300)
        self.update_gui()
        self.show()

    def on_next_clicked(self):
        current_index_int = self.wizard_qsw_w3.currentIndex()
        if current_index_int >= self.wizard_qsw_w3.count() - 1:
            self.close_signal.emit(True)
            self.close()

        logging.debug("current_index_int = " + str(current_index_int))
        self.wizard_qsw_w3.setCurrentIndex(current_index_int + 1)
        self.update_gui()

    def on_prev_clicked(self):
        current_index_int = self.wizard_qsw_w3.currentIndex()
        if current_index_int <= 0:
            return
        logging.debug("current_index_int = " + str(current_index_int))
        self.wizard_qsw_w3.setCurrentIndex(current_index_int - 1)
        self.update_gui()

    def update_gui(self):
        current_index_int = self.wizard_qsw_w3.currentIndex()
        self.prev_qpb.setDisabled(current_index_int == 0)

        if current_index_int == self.wizard_qsw_w3.count() - 1:
            self.next_qpb.setText("Finish")  # "open breathing dialog"
        else:
            self.next_qpb.setText(NEXT)

        if self.wizard_qsw_w3.currentWidget() == self.br_dlg_page:
            self.breathing_dlg.setFocus()


class IntroPage(QtWidgets.QWidget):
    def __init__(self, i_title: str, *i_widgets):
        super().__init__()
        self.setSizePolicy(
            self.sizePolicy().horizontalPolicy(),
            QtWidgets.QSizePolicy.Maximum
        )

        self.vbox_l2 = QtWidgets.QVBoxLayout()
        (cm_left, cm_top, cm_right, cm_bottom) = self.vbox_l2.getContentsMargins()
        self.vbox_l2.setContentsMargins(40, cm_top, 20, cm_bottom)
        self.setLayout(self.vbox_l2)
        self.vbox_l2.addSpacing(MARGIN_TOP)
        self.title_qll = QtWidgets.QLabel(i_title)
        # self.title_qll.setTextFormat
        # self.title_qll.setAlignment(QtCore.Qt.AlignHCenter)
        self.title_qll.setFont(matc.globa.get_font(matc.globa.FontSize.xlarge))
        self.vbox_l2.addWidget(self.title_qll)
        self.vbox_l2.addSpacing(WIDGET_SPACING)

        for widget in i_widgets:
            self.vbox_l2.addWidget(widget)
            self.vbox_l2.addSpacing(WIDGET_SPACING)

        self.vbox_l2.addSpacing(WIDGET_SPACING)
        self.vbox_l2.addStretch(1)



import logging

# Dependencies:
import svgutils  # https://svgutils.readthedocs.io/en/latest/
import cairosvg

"""
online search: inkscape save only shape not page

rotate svg
https://stackoverflow.com/questions/43199869/rotate-and-scale-a-complete-svg-document-using-python
"""

FILE_NAME = "moving-icon-gradient.svg"
orig_svg_figure = svgutils.transform.fromfile(FILE_NAME)
assert orig_svg_figure.width == orig_svg_figure.height
print(f"{orig_svg_figure.width=} (height is same)")

index = 0
numeric = "0123456789."
for char in orig_svg_figure.width:
    if char not in numeric:
        break
    index += 1
width_and_height: float = float(orig_svg_figure.width[:index])

orig_svg = svgutils.compose.SVG(FILE_NAME)
orig_svg.rotate(60, width_and_height/2, width_and_height/2)
# https://svgutils.readthedocs.io/en/latest/transform.html#svgutils.transform.FigureElement.rotate
new_svg_figure = svgutils.compose.Figure(orig_svg_figure.height, orig_svg_figure.width, orig_svg)
new_svg_figure.save("test-rotate-60.svg")

cairosvg.svg2png(
    bytestring=open("test-rotate-60.svg").read().encode('utf-8'),
    write_to="test-rotate-60.png"
)



from PyQt5 import QtWidgets, QtCore
from matc import globa


class H1(QtWidgets.QLabel):
    """
    Heading 1
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class H2(QtWidgets.QLabel):
    """
    Heading 2
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class HorizontalLine(QtWidgets.QFrame):
    def __init__(self, *__args):
        super().__init__(*__args)
        self.setFrameShape(self.HLine)


class RadioButtonLeft(QtWidgets.QRadioButton):
    """
    The left button of a radio button group
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class RadioButtonMiddle(QtWidgets.QRadioButton):
    """
    One of the middle buttons of a radio button group
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class RadioButtonRight(QtWidgets.QRadioButton):
    """
    The right button of a radio button group
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class PushButton(QtWidgets.QPushButton):
    """
    A rectangular button
    """
    def __init__(self, *__args):
        super().__init__(*__args)


class PhrasesList(QtWidgets.QListWidget):
    def __init__(self, *__args):
        super().__init__(*__args)


class PageGrid(QtWidgets.QGridLayout):
    def __init__(self, *__args):
        super().__init__(*__args)
        if QtCore.QSysInfo.kernelType() != "darwin":
            self.setVerticalSpacing(globa.GRID_VERTICAL_SPACING_LINUX)


class ButtonGrid(QtWidgets.QGridLayout):
    def __init__(self, *__args):
        super().__init__(*__args)
        if QtCore.QSysInfo.kernelType() != "darwin":
            self.setHorizontalSpacing(globa.BUTTON_BAR_HORIZONTAL_SPACING_LINUX)



import datetime
import json
import logging
import os
import shutil
import matc.globa

JSON_OBJ_TYPE = "__obj_type__"

# Setting Keys (SK)
SK_SHOW_BREATHING_TEXT = "show_breathing_text"
SK_BREATHING_AUDIO_VOLUME = "breathing_audio_volume"
SK_BREATHING_BREAK_TIMER_SECS = "breathing_break_timer_secs"
SK_BREATHING_PHRASES = "breathing_phrases"
SK_BREATHING_VISUALIZATION = "breathing_visualization"
SK_MOVE_MOUSE_CURSOR = "move_mouse_cursor"
# SK_DEBUG_USE_SECONDS_FOR_BR_TIMER = "debug_use_seconds_for_br_timer"

"""
SK_NR_OF_TIMES_UNTIL_FEEDBACK_SHOWN = "nr_of_times_until_feedback_shown"
SK_NR_OF_TIMES_UNTIL_FEEDBACK_SHOWN: matc.globa.INITIAL_NR_OF_TIMES_UNTIL_FEEDBACK_SHOWN,
FEEDBACK_DIALOG_NOT_SHOWN_AT_STARTUP = -1
INITIAL_NR_OF_TIMES_UNTIL_FEEDBACK_SHOWN = 10

    def update_gui(self):
        self.gui_update_bool = True
        # settings = matc.settings.settings.is.SettingsM.get()
        self.show_again_qcb.setChecked(
            settings.nr_times_started_since_last_feedback_notif != matc.globa.FEEDBACK_DIALOG_NOT_SHOWN_AT_STARTUP
        )
        self.gui_update_bool = False

    def on_show_again_toggled(self, i_checked: bool):
        if self.gui_update_bool:
            return
        settings = matc.model.SettingsM.get()
        if i_checked:
            if settings.nr_times_started_since_last_feedback_notif == matc.globa.FEEDBACK_DIALOG_NOT_SHOWN_AT_STARTUP:
                settings.nr_times_started_since_last_feedback_notif = 0
            else:
                pass
        else:
            settings.nr_times_started_since_last_feedback_notif = matc.globa.FEEDBACK_DIALOG_NOT_SHOWN_AT_STARTUP


    self.show_again_qcb = QtWidgets.QCheckBox(self.tr("Show this dialog at startup again in the future"))
    self.show_again_qcb.toggled.connect(self.on_show_again_toggled)
    vbox_l2.addWidget(self.show_again_qcb)


"""

BREATHING_BREAK_TIMER_DEFAULT_SECS = 540

settings: dict = {
    SK_SHOW_BREATHING_TEXT: True,
    SK_BREATHING_AUDIO_VOLUME: 50,
    SK_BREATHING_BREAK_TIMER_SECS: BREATHING_BREAK_TIMER_DEFAULT_SECS,
    SK_BREATHING_PHRASES: [],
    SK_BREATHING_VISUALIZATION: matc.globa.BreathingVisalization.bar.value,
    SK_MOVE_MOUSE_CURSOR: False,
}
# -the values given here are the minimum values needed for the application to work
# settings_dict[SETTING_ONE_KEY]


# @dataclasses.dataclass
# In the future we may want to rewrite the three classes below to be dataclasses instead
# https://docs.python.org/3/library/dataclasses.html
# They are supported from Python 3.7

class SettingsListObject:
    def __init__(self, i_id: int):
        self.id: int = i_id


class BreathingPhrase(SettingsListObject):
    def __init__(self, i_id: int, i_in_breath: str, i_out_breath: str):
        super().__init__(i_id)
        self.in_breath: str = i_in_breath
        self.out_breath: str = i_out_breath


class RestAction(SettingsListObject):
    def __init__(self, i_id: int, i_title: str, i_image_path: str):
        super().__init__(i_id)
        self.title: str = i_title
        self.image_path: str = i_image_path


def _get_list_object(settings_key: str, i_list_id: int):
    # -> SettingsListObject
    # SettingsListObject
    list_objects: list = settings[settings_key]
    for o in list_objects:
        if o.id == i_list_id:
            return o
    raise Exception(f"No list object found in the list {settings_key} for the id {i_list_id}")


def get_topmost_breathing_phrase() -> BreathingPhrase:
    list_objects: list = settings[SK_BREATHING_PHRASES]
    if len(list_objects) < 1:
        raise Exception("List is empty, so cannot return item")
    return list_objects[0]


def get_breathing_phrase(i_id: int) -> BreathingPhrase:
    return _get_list_object(SK_BREATHING_PHRASES, i_id)


def get_breathing_volume() -> int:
    return settings[SK_BREATHING_AUDIO_VOLUME]


def _add_list_object(i_settings_key: str, i_class, *args) -> int:
    """
    Order of JSON arrays is preserved (and of course Python lists too)
    https://stackoverflow.com/a/7214312/2525237
    """
    list_objects: list = settings[i_settings_key]
    highest_id: int = 0
    if list_objects:
        highest_id: int = max(lo.id for lo in list_objects)
    new_id: int = highest_id + 1
    new_br_phrase = i_class(new_id, *args)
    list_objects.append(new_br_phrase)
    # settings[i_settings_key] = list_objects
    return new_id


def add_breathing_phrase(i_in_breath: str, i_out_breath: str) -> int:
    new_id = _add_list_object(SK_BREATHING_PHRASES, BreathingPhrase, i_in_breath, i_out_breath)
    return new_id


def _remove_list_object(i_settings_key: str, i_id: int) -> None:
    list_objects: list = settings[i_settings_key]
    for o in list_objects:
        if o.id == i_id:
            list_objects.remove(o)
            return


def remove_breathing_phrase(i_id: int):
    _remove_list_object(SK_BREATHING_PHRASES, i_id)


def _set_list_object_attributes(i_settings_key: str, i_id: int, **kwargs):
    list_objects: list = settings[i_settings_key]
    for o in list_objects:
        if o.id == i_id:
            for k, v in kwargs.items():
                if getattr(o, k):
                    setattr(o, k, v)
                else:
                    logging.warning(f"Could not find attribute {k} in object {o}. Continuing")
            return


def set_breathing_phrase_attributes(i_id: int, **kwargs):
    # Example: set_breathing_phrase_attributes(1, in_breath="Breathing in, i know i ___")
    # Changing to these arguments? in_breath=i_in_breath, out_breath=i_out_breath
    _set_list_object_attributes(SK_BREATHING_PHRASES, i_id, **kwargs)


class MyEncoder(json.JSONEncoder):
    def default(self, obj):  # -overridden
        if issubclass(type(obj), SettingsListObject):
            object_dictionary: dict = obj.__dict__
            type_value = type(obj).__name__
            # if isinstance(obj, BreathingPhrase):
            # type_value = BreathingPhrase.__name__
            # else: raise Exception(f"Cannot endode object: Case is not covered")
            object_dictionary[JSON_OBJ_TYPE] = type_value
            return obj.__dict__
        else:
            return super().default(obj)


def my_decode(dct: dict):
    """
    :param dct:
    From the documentation:
    "object_hook is an optional function that will be called with the result of any object literal
    decoded (a dict). The return value of object_hook will be used instead of the dict."
    :return:
    """
    if JSON_OBJ_TYPE in dct and dct[JSON_OBJ_TYPE] == RestAction.__name__:
        rest_action_obj = RestAction(
            i_id=dct["id"],
            i_title=dct["title"],
            i_image_path=dct["image_path"]
        )
        return rest_action_obj
    if JSON_OBJ_TYPE in dct and dct[JSON_OBJ_TYPE] == BreathingPhrase.__name__:
        breathing_phrase_obj = BreathingPhrase(
            i_id=dct["id"],
            i_in_breath=dct["in_breath"],
            i_out_breath=dct["out_breath"]
        )
        return breathing_phrase_obj
    return dct


def settings_file_exists() -> bool:
    return os.path.isfile(matc.globa.settings_file_path)


def backup_settings_file() -> None:
    if matc.globa.testing_bool:
        return
    date_sg = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    new_file_name_sg = matc.globa.get_settings_file_path(date_sg)
    shutil.copyfile(matc.globa.settings_file_path, new_file_name_sg)

    # Removing older backups

    # Checking if it's well-formatted (JSON ok?)


def save_settings_to_json_file():
    logging.debug("Saving to json file")
    logging.debug(f"{matc.settings.settings=}")
    with open(matc.globa.settings_file_path, "w") as write_file:
        json.dump(matc.settings.settings, write_file, indent=2, cls=MyEncoder)


def update_settings_dict_with_json_data(i_min_dict_ref: dict, i_file_path: str):
    logging.debug("update_dict_with_json_data")
    if os.path.isfile(i_file_path):
        with open(i_file_path, "r") as read_file:
            # try:
            #     pass
            # except json.decoder.JSONDecodeError:
            #     pass

            from_file_dict: dict = json.load(read_file, object_hook=my_decode)

            diff_key_list: list = []
            for min_key in i_min_dict_ref.keys():
                if min_key not in from_file_dict.keys():
                    diff_key_list.append(min_key)
            if diff_key_list:
                # diff_keys_str = ", ".join(diff_key_list)
                logging.warning(f"One or more keys needed for the application to work were not "
                f"available in {os.path.basename(i_file_path)} so have been added now "
                f"(with a default value). These are the keys: {diff_key_list}")

            diff_key_list: list = []
            for file_key in from_file_dict.keys():
                if file_key not in i_min_dict_ref.keys():
                    diff_key_list.append(file_key)
            if diff_key_list:
                # diff_keys_str = ", ".join(diff_key_list)
                logging.warning(f"One or more keys in {os.path.basename(i_file_path)} are not "
                f"used by the application (though may have been used before). "
                f"These are the keys: {diff_key_list}")

            for min_key in i_min_dict_ref.keys():
                if min_key not in from_file_dict.keys():
                    diff_key_list.append(min_key)
            for file_key in from_file_dict.keys():
                if file_key not in i_min_dict_ref.keys():
                    diff_key_list.append(file_key)

            # logging.debug(f"Before merge {i_min_dict_ref=}")
            # logging.debug(f"Before merge {from_file_dict=}")
            i_min_dict_ref.update(from_file_dict)
            # -if there are different values for the same key, the value
            #  in from_file_dict takes precendence
            # logging.debug(f"After merge {i_min_dict_ref=}")


# Initial setup
if not settings_file_exists():
    # min_settings_dict[SK_REST_ACTIONS].update(init_rest_actions)
    add_breathing_phrase(
        "Breathing in I know I am breathing in",
        "Breathing out I know I am breathing out"
    )
    add_breathing_phrase(
        "Breathing in I follow the whole length of my in-breath",
        "Breathing out I follow the whole length of my out-breath"
    )
    add_breathing_phrase(
        "Breathing in I am aware of my body",
        "Breathing out I am aware of my body"
    )
    add_breathing_phrase(
        "Breathing in I am aware of my posture",
        "Breathing out I adjust my posture"
    )
    add_breathing_phrase(
        "May everyone live with compassion",
        "May everyone live in peace"
    )
    """
    "Breathing in, I know I am breathing in","Breathing out, I know I am breathing out",
    "Aware of my body, I breathe in","Aware of my body, I breathe out",
    "Breathing in, I care for my body","Breathing out, I relax my body",
    "Happy, At Peace","May I be happy",
    "Breathing in I share the well-being of others","Breathing out I contribute to the well-being of others",
    "Breathing in compassion to myself","Breathing out compassion to others",
    "Self-love and acceptance","I love and accept myself just as I am",
    """


# update_dict_with_json_data(settings, matc.globa.get_settings_file_path())
# logging.debug(f"{settings=}")



import sys
import unittest
from PySide6 import QtTest
from PySide6 import QtCore
from PySide6 import QtWidgets
import matc.main_object
import matc.globa


test_app = QtWidgets.QApplication(sys.argv)
# -has to be set here (rather than in __main__) to avoid an error


# https://doc.qt.io/qt-5/qpluginloader.html

# https://doc.qt.io/qt-5/qlibraryinfo.html
plugins_path: str = QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.PluginsPath)
print(f"{plugins_path=}")


class MainTest(unittest.TestCase):
    """
    "@unittest.skip" can be used to skip a test
    """

    @classmethod
    def setUpClass(cls):
        matc.globa.testing_bool = True

    def setUp(self):
        pass

    def test_main_object(self):
        i = 0


if __name__ == "__main__":
    unittest.main()



import logging
import logging.handlers
import os
import sys
import argparse
from PyQt5 import QtCore
from PyQt5 import QtWidgets
import matc.constants
import matc.globa
import matc.settings
import matc.main_object

# The following import looks like it isn't used, but it is necessary for importing the images.
import matc.matc_rc  # pylint: disable=unused-import

LOG_FILE_NAME_STR = "matc.log"


def on_about_to_quit():
    logging.debug("on_about_to_quit --- saving settings to json file (in the user config dir)")
    matc.settings.save_settings_to_json_file()


def main():
    # db_filepath: str = matc.globa.get_database_path()
    # matc.globa.db_file_exists_at_application_startup_bl = os.path.isfile(db_filepath)
    # -settings this variable before the file has been created

    logger = logging.getLogger()
    # -if we set a name here for the logger the file handler will no longer work, unknown why
    logger.handlers = []  # -removing the default stream handler first
    # logger.propagate = False
    log_path_str = matc.globa.get_config_path(LOG_FILE_NAME_STR)
    rfile_handler = logging.handlers.RotatingFileHandler(log_path_str, maxBytes=8192, backupCount=2)
    rfile_handler.setLevel(logging.WARNING)
    # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    rfile_handler.setFormatter(formatter)
    logger.addHandler(rfile_handler)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    default_settings_file_path = matc.globa.get_settings_file_path()
    argument_parser = argparse.ArgumentParser(prog=matc.constants.APPLICATION_PRETTY_NAME,
        description=matc.constants.SHORT_DESCR_STR)
    argument_parser.add_argument("--settings-file", "-s", default=default_settings_file_path,
        help="Path to a settings file (json format). If only a file name is given, the default directory will be used")
    parsed_args = argument_parser.parse_args()
    dir_ = os.path.dirname(parsed_args.settings_file)
    if dir_ and os.path.isdir(dir_):
        matc.globa.settings_file_path = parsed_args.settings_file
    else:
        default_dir = os.path.dirname(default_settings_file_path)
        matc.globa.settings_file_path = os.path.join(default_dir, parsed_args.settings_file)
    matc.settings.update_settings_dict_with_json_data(matc.settings.settings, matc.globa.settings_file_path)

    matc_qapplication = QtWidgets.QApplication(sys.argv)

    # Application information
    matc.globa.sys_info_telist.append(("Application name", matc.constants.APPLICATION_NAME))
    matc.globa.sys_info_telist.append(("Application version", matc.constants.APPLICATION_VERSION))
    matc.globa.sys_info_telist.append(("Config path", matc.globa.get_config_path()))
    matc.globa.sys_info_telist.append(("Module path", matc.globa.get_module_path()))
    matc.globa.sys_info_telist.append(("Python version", sys.version))
    matc.globa.sys_info_telist.append(("Qt version", QtCore.qVersion()))
    matc.globa.sys_info_telist.append(("PyQt (Python module) version", QtCore.PYQT_VERSION_STR))
    sys_info = QtCore.QSysInfo()
    matc.globa.sys_info_telist.append(("OS name and version", sys_info.prettyProductName()))
    matc.globa.sys_info_telist.append(
        ("Kernel type and version", sys_info.kernelType() + " " + sys_info.kernelVersion()))
    matc.globa.sys_info_telist.append(("buildCpuArchitecture", sys_info.buildCpuArchitecture()))
    matc.globa.sys_info_telist.append(("currentCpuArchitecture", sys_info.currentCpuArchitecture()))

    # set stylesheet
    stream = QtCore.QFile(os.path.join(matc.globa.get_module_path(), "matc.qss"))
    stream.open(QtCore.QIODevice.ReadOnly)
    matc_qapplication.setStyleSheet(QtCore.QTextStream(stream).readAll())

    # desktop_widget = matc_qapplication.desktop()
    # matc.globa.sys_info_telist.append(("Virtual desktop", str(desktop_widget.isVirtualDesktop())))
    # matc.globa.sys_info_telist.append(("Screen count", str(desktop_widget.screenCount())))
    # matc.globa.sys_info_telist.append(("Primary screen", str(desktop_widget.primaryScreen())))

    system_locale = QtCore.QLocale.system().name()
    logging.info('System Localization: ' + system_locale)
    matc_qapplication.setQuitOnLastWindowClosed(False)
    matc_qapplication.aboutToQuit.connect(on_about_to_quit)

    matc_main_object = matc.main_object.MainObject()

    sys.exit(matc_qapplication.exec_())


if __name__ == "__main__":
    main()



APPLICATION_PRETTY_NAME = "Mindfulness at the Computer"
APPLICATION_NAME = "mindfulness-at-the-computer"
APPLICATION_VERSION = "1.0.0-alpha.9"
SHORT_DESCR_STR = "Helps you stay mindful of your breathing while using your computer."



from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install
# -there's no setuptools.command.uninstall (or similar)
import os
import matc.constants

long_description_str = ""
this_dir_abs_path_str = os.path.dirname(__file__)
readme_abs_path_str = os.path.join(this_dir_abs_path_str, "README.md")


class PostDevelopCommand(develop):
    def run(self):
        develop.run(self)
        import extra_setup
        extra_setup.do_extra_setup()


class PostInstallCommand(install):
    def run(self):
        install.run(self)
        import extra_setup
        extra_setup.do_extra_setup()


try:
    with open(readme_abs_path_str, "r") as file:
        long_description_str = '\n' + file.read()
except FileNotFoundError:
    long_description_str = matc.constants.SHORT_DESCR_STR

setup(
    name=matc.constants.APPLICATION_NAME,
    version=matc.constants.APPLICATION_VERSION,
    packages=['matc', 'matc.gui'],
    url="https://mindfulness-at-the-computer.gitlab.io",
    license='GPLv3',
    author='Tord Dellsén, and others',
    author_email='tord.dellsen@gmail.com',
    description=matc.constants.SHORT_DESCR_STR,
    include_package_data=True,
    install_requires=["PyQt5>=5.15.2"],
    entry_points={"console_scripts": [f"{matc.constants.APPLICATION_NAME}=matc.main:main"]},
    long_description_content_type='text/markdown',
    long_description=long_description_str,
    python_requires='>=3.6.0',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Other/Nonlisted Topic'
    ],
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand
    }
)
"""
Ubuntu versions and Python versions:
18.04 LTS: 3.6 - f-strings,
3.7 - 
20.04 LTS: 3.8 - 
21.04: 3.9 - 
22.04 - 3.10 - 

To install earlier versions:
https://www.digitalocean.com/community/questions/how-to-install-a-specific-python-version-on-ubuntu

https://www.python.org/downloads/
tar xzvf Python-3.5.0.tgz
cd Python-3.5.0
./configure
make
sudo make install
https://askubuntu.com/a/727814/360991

sudo add-apt-repository ppa:deadsnakes/ppa
sudo apt-get update
sudo apt-get install python3.6
https://askubuntu.com/a/682875/360991
Doesn't work for 3.6

List of classifiers:
https://pypi.org/classifiers/

"""



import logging
import enum
import sys
import webbrowser
import os.path
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import matc.constants
import matc.globa
import matc.settings
import matc.gui.intro_dlg
# import matc.gui.breathing_phrase_list_wt

NEW_ROW: int = -1


class SysinfoDialog(QtWidgets.QDialog):
    def __init__(self, i_parent=None):
        super().__init__(i_parent)
        self.setModal(True)

        vbox_l2 = QtWidgets.QVBoxLayout(self)

        self._system_info_str = '\n'.join([
            descr_str + ": " + str(value) for (descr_str, value) in matc.globa.sys_info_telist
        ])

        self.system_info_qll = QtWidgets.QLabel(self._system_info_str)
        vbox_l2.addWidget(self.system_info_qll)

        self.copy_qpb = QtWidgets.QPushButton(self.tr("Copy to clipboard"))
        self.copy_qpb.clicked.connect(self.on_copy_button_clicked)
        vbox_l2.addWidget(self.copy_qpb)

        self.button_box = QtWidgets.QDialogButtonBox(
            QtCore.Qt.Horizontal,
            self
        )
        self.button_box.setStandardButtons(QtWidgets.QDialogButtonBox.Close)

        vbox_l2.addWidget(self.button_box)
        self.button_box.rejected.connect(self.reject)
        # -accepted and rejected are "slots" built into Qt

    def on_copy_button_clicked(self):
        qclipboard = QtGui.QGuiApplication.clipboard()
        qclipboard.setText(self._system_info_str)
        # -this will copy the text to the system clipboard


class FeedbackDialog(QtWidgets.QDialog):

    def __init__(self):
        super().__init__()
        self.gui_update_bool = True

        vbox_l2 = QtWidgets.QVBoxLayout()
        self.setLayout(vbox_l2)

        help_request_str = """<h3>Help Us</h3>
<p>We are grateful for feedback, for example please contact us if you</p>
<ul>
<li>find a bug</li>
<li>have a suggestion for a new feature</li>
<li>have ideas for how to improve the interface</li>
<li>have feedback about what you like about the application and how it helps you when using the computer (we are looking for testimonials!)</li>
</ul>
<p>You can reach us using this email address:</p>"""

        help_request_qll = QtWidgets.QLabel()
        # help_request_qll.setFont(matc.globa.get_font_large())
        help_request_qll.setText(help_request_str)
        help_request_qll.setWordWrap(True)
        vbox_l2.addWidget(help_request_qll)

        email_qll = QtWidgets.QLabel()
        email_qll.setFont(matc.globa.get_font(matc.globa.FontSize.xxlarge))
        email_qll.setText('sunyata.software@gmail.com')
        vbox_l2.addWidget(email_qll)

        emailus_qpb = QtWidgets.QPushButton("Email us!")
        emailus_qpb.clicked.connect(self.on_emailus_clicked)
        vbox_l2.addWidget(emailus_qpb)

        self.show()

    def on_emailus_clicked(self):
        url_string = "mailto:sunyata.software@gmail.com?subject=Feedback"
        webbrowser.open(url_string)
        # Alt: QtGui.QDesktopServices.openUrl(QtCore.QUrl(url_string))


class ToggleSwitchWt(QtWidgets.QWidget):
    toggled_signal = QtCore.pyqtSignal(bool)

    def __init__(self):
        super().__init__()

        self.updating_gui_bool = False

        self.turn_on_off_qcb = QtWidgets.QCheckBox()
        self.turn_on_off_qcb.toggled.connect(self._on_toggled)
        on_off_qhl = QtWidgets.QHBoxLayout()
        on_off_qhl.setContentsMargins(0, 0, 0, 0)
        on_off_qhl.addWidget(QtWidgets.QLabel(self.tr("Turn the dialog and notifications on or off")))
        on_off_qhl.addStretch(1)
        on_off_qhl.addWidget(self.turn_on_off_qcb)
        self.setLayout(on_off_qhl)

    def _on_toggled(self, i_checked: bool):
        if self.updating_gui_bool:
            return
        self.toggled_signal.emit(i_checked)

    def update_gui(self, i_checked: bool):
        self.updating_gui_bool = True

        self.turn_on_off_qcb.setChecked(i_checked)

        self.updating_gui_bool = False


class MoveDirectionEnum(enum.Enum):
    up = 1
    down = 2


class BreathingPhraseEditDialog(QtWidgets.QDialog):
    def __init__(self, i_id: int, i_parent=None):
        super().__init__(parent=i_parent)
        self.setModal(True)
        self.setMinimumWidth(250)
        vbox = QtWidgets.QVBoxLayout(self)

        """
        # If a phrase is not selected, default to phrase with id 1
        if matc.globa.active_phrase_id_it == matc.globa.NO_PHRASE_SELECTED_INT:
            matc.globa.active_phrase_id_it = 1
        """

        bp_obj = matc.settings.get_breathing_phrase(i_id)

        self.breath_title_qle = QtWidgets.QLineEdit(str(bp_obj.id))
        vbox.addWidget(QtWidgets.QLabel(self.tr("ID")))
        vbox.addWidget(self.breath_title_qle)

        vbox.addWidget(QtWidgets.QLabel(self.tr("Phrase")))
        self.in_breath_phrase_qle = QtWidgets.QLineEdit(bp_obj.in_breath)
        vbox.addWidget(self.in_breath_phrase_qle)

        self.out_breath_phrase_qle = QtWidgets.QLineEdit(bp_obj.out_breath)
        vbox.addWidget(self.out_breath_phrase_qle)

        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal,
            self
        )
        vbox.addWidget(self.button_box)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        # -accept and reject are "slots" built into Qt

        self.adjustSize()

    @staticmethod
    def start(i_id) -> bool:
        dlg = BreathingPhraseEditDialog(i_id)
        dlg.exec()

        if dlg.result() == QtWidgets.QDialog.Accepted:
            logging.debug("dlg.result() == QtWidgets.QDialog.Accepted")
            bp_obj = matc.settings.get_breathing_phrase(i_id)
            bp_obj.in_breath = dlg.in_breath_phrase_qle.text()
            bp_obj.out_breath = dlg.out_breath_phrase_qle.text()
            return True
        return False


class MyListWidget(QtWidgets.QListWidget):
    drop_signal = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)

    def dropEvent(self, QDropEvent):
        super().dropEvent(QDropEvent)
        self.drop_signal.emit()
        # self.update_db_sort_order_for_all_rows()


class SettingsWin(QtWidgets.QMainWindow):
    show_intro_dialog_signal = QtCore.pyqtSignal()
    br_timer_change_signal = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.setGeometry(100, 64, 400, 540)
        self.setWindowTitle(f"Settings Dialog - {matc.constants.APPLICATION_PRETTY_NAME}")
        self.setWindowIcon(QtGui.QIcon(matc.globa.get_app_icon_path("icon.png")))

        self.updating_gui: bool = True
        self.widget_w1 = QtWidgets.QWidget()
        self.seconds_for_br_timer_action = None

        """
        if matc.globa.testing_bool:
            data_storage_str = "{Testing - data stored in memory}"
        else:
            data_storage_str = "{Live - data stored on hard drive}"
        window_title_str = (
                matc.constants.APPLICATION_PRETTY_NAME
                + " [" + matc.constants.APPLICATION_VERSION + "] "
                + data_storage_str
        )
        self.setWindowTitle(window_title_str)
        """

        self.setStyleSheet(
            "selection-background-color:" + matc.globa.LIGHT_GREEN_COLOR + ";"
            "selection-color:#000000;"
        )

        self.setCentralWidget(self.widget_w1)

        # Setup of Menu
        self.menu_bar = self.menuBar()
        self.update_menu()


        vbox_l2 = QtWidgets.QVBoxLayout()
        self.widget_w1.setLayout(vbox_l2)

        hbox_br_time_l3 = QtWidgets.QHBoxLayout()
        vbox_l2.addLayout(hbox_br_time_l3)
        hbox_br_time_l3.addWidget(QtWidgets.QLabel("Breathing break time"))
        self.breathing_break_time_qsb = QtWidgets.QSpinBox()
        self.breathing_break_time_qsb.setMinimum(1)
        self.breathing_break_time_qsb.setMaximum(1000)
        hbox_br_time_l3.addWidget(self.breathing_break_time_qsb)
        self.breathing_break_time_qsb.valueChanged.connect(self.on_br_time_value_changed)
        hbox_br_time_l3.addWidget(QtWidgets.QLabel("minutes"))

        hbox_volume_l3 = QtWidgets.QHBoxLayout()
        vbox_l2.addLayout(hbox_volume_l3)
        hbox_volume_l3.addWidget(QtWidgets.QLabel("Volume"))
        self.volume_qsr = QtWidgets.QSlider()
        hbox_volume_l3.addWidget(self.volume_qsr)
        self.volume_qsr.valueChanged.connect(self.on_volume_changed)
        self.volume_qsr.setMinimum(0)
        self.volume_qsr.setMaximum(100)
        self.volume_qsr.setOrientation(QtCore.Qt.Horizontal)

        self.phrases_qgb = QtWidgets.QGroupBox("Breathing phrases")
        vbox_l2.addWidget(self.phrases_qgb)
        vbox_l4 = QtWidgets.QVBoxLayout()
        self.phrases_qgb.setLayout(vbox_l4)

        self.edit_bp_qpb = QtWidgets.QPushButton("Edit")

        self.breathing_phrases_qlw = MyListWidget()
        vbox_l4.addWidget(self.breathing_phrases_qlw)
        self.breathing_phrases_qlw.setSpacing(2)
        self.breathing_phrases_qlw.currentRowChanged.connect(self.on_bp_current_row_changed)
        self.breathing_phrases_qlw.itemDoubleClicked.connect(self.on_bp_double_clicked)
        self.breathing_phrases_qlw.drop_signal.connect(self.on_bp_item_dropped)
        self.populate_bp_list()

        hbox_buttons_l5 = QtWidgets.QHBoxLayout()
        vbox_l4.addLayout(hbox_buttons_l5)

        hbox_buttons_l5.addWidget(self.edit_bp_qpb)
        self.edit_bp_qpb.clicked.connect(self.on_edit_bp_clicked)

        self.add_bp_qpb = QtWidgets.QPushButton("Add")
        hbox_buttons_l5.addWidget(self.add_bp_qpb)
        self.add_bp_qpb.clicked.connect(self.on_add_bp_clicked)

        self.del_bp_qpb = QtWidgets.QPushButton("Del")
        hbox_buttons_l5.addWidget(self.del_bp_qpb)
        self.del_bp_qpb.clicked.connect(self.on_del_bp_clicked)

        self.set_active_bp_qpb = QtWidgets.QPushButton("Set Active")
        hbox_buttons_l5.addWidget(self.set_active_bp_qpb)
        self.set_active_bp_qpb.clicked.connect(self.on_set_active_bp_clicked)

        self.visualizations_qgb = QtWidgets.QGroupBox("Breathing visualizations")
        vbox_l2.addWidget(self.visualizations_qgb)
        vbox_l4 = QtWidgets.QVBoxLayout()
        self.visualizations_qgb.setLayout(vbox_l4)
        self.br_vis_group_qbg = QtWidgets.QButtonGroup()
        self.br_vis_bar_qrb = QtWidgets.QRadioButton("Bar")
        self.br_vis_group_qbg.addButton(self.br_vis_bar_qrb, matc.globa.BreathingVisalization.bar.value)
        vbox_l4.addWidget(self.br_vis_bar_qrb)
        self.br_vis_circle_qrb = QtWidgets.QRadioButton("Circle")
        self.br_vis_group_qbg.addButton(self.br_vis_circle_qrb, matc.globa.BreathingVisalization.circle.value)
        vbox_l4.addWidget(self.br_vis_circle_qrb)
        self.br_vis_group_qbg.idClicked.connect(self.on_br_vis_id_clicked)

        self.move_mouse_cursor_qcb = QtWidgets.QCheckBox("Move mouse cursor to breathing dialog")
        vbox_l2.addWidget(self.move_mouse_cursor_qcb)
        self.move_mouse_cursor_qcb.clicked.connect(self.on_move_mouse_cursor_clicked)
        self.move_mouse_cursor_qcb.setToolTip("Useful if you are using a touchpad")

        ######## self.show_breathing_phrase_qcb = QtWidgets.QCheckBox("Show breahing phrase")
        ######## vbox_l4.addWidget(self.show_breathing_phrase_qcb)
        ######## self.show_breathing_phrase_qcb.toggled.connect(self.on_show_br_phrase_toggled)

        # matc.globa.active_phrase_id_it

        self.update_gui()

    def populate_bp_list(self):
        self.breathing_phrases_qlw.clear()
        phrases: list[matc.settings.BreathingPhrase] = matc.settings.settings[matc.settings.SK_BREATHING_PHRASES]
        for p in phrases:
            self._add_bp_to_gui(p.id)

    def on_bp_item_dropped(self):
        new_order_phrase_list = []
        for item_row in range(self.breathing_phrases_qlw.count()):
            item = self.breathing_phrases_qlw.item(item_row)
            item_id = item.data(QtCore.Qt.UserRole)
            phrase = matc.settings.get_breathing_phrase(item_id)
            new_order_phrase_list.append(phrase)
        matc.settings.settings[matc.settings.SK_BREATHING_PHRASES] = new_order_phrase_list
        # self.populate_bp_list()

    def on_bp_double_clicked(self, i_item: QtWidgets.QListWidgetItem):
        matc.globa.active_phrase_id = i_item.data(QtCore.Qt.UserRole)
        self.update_gui_active_bold()

    def on_move_mouse_cursor_clicked(self, i_checked: bool):
        if self.updating_gui:
            return
        matc.settings.settings[matc.settings.SK_MOVE_MOUSE_CURSOR] = i_checked

    def on_br_vis_id_clicked(self, i_id: int):
        if self.updating_gui:
            return
        matc.settings.settings[matc.settings.SK_BREATHING_VISUALIZATION] = i_id

    def on_bp_current_row_changed(self):
        pass

    def on_show_br_phrase_toggled(self, i_checked: bool):
        logging.debug("on_show_br_phrase_toggled")
        if i_checked:
            top_qlwi = self.breathing_phrases_qlw.item(0)
            matc.globa.active_phrase_id = top_qlwi.data(QtCore.Qt.UserRole)
        else:
            matc.globa.active_phrase_id = matc.globa.BREATHING_PHRASE_NOT_SET
        self.update_gui_active_bold()

    def on_set_active_bp_clicked(self):
        current_qlwi = self.breathing_phrases_qlw.currentItem()
        matc.globa.active_phrase_id = current_qlwi.data(QtCore.Qt.UserRole)
        self.update_gui_active_bold()

    def update_gui_active_bold(self):
        for item_row in range(self.breathing_phrases_qlw.count()):
            item = self.breathing_phrases_qlw.item(item_row)
            item_font = item.font()
            if item.data(QtCore.Qt.UserRole) == matc.globa.active_phrase_id:
                item_font.setBold(True)
                item.setFont(item_font)
            elif item_font.bold():
                item_font.setBold(False)
                item.setFont(item_font)
            """
            self.show_breathing_phrase_qcb.setChecked(
                matc.globa.active_phrase_id != matc.globa.BREATHING_PHRASE_NOT_SHOWN
            )
            """

    def on_volume_changed(self, i_new_value: int):
        if self.updating_gui:
            return
        matc.settings.settings[matc.settings.SK_BREATHING_AUDIO_VOLUME] = i_new_value

    def on_add_bp_clicked(self):
        new_id: int = matc.settings.add_breathing_phrase("ib", "ob")
        result = BreathingPhraseEditDialog.start(new_id)
        if result:
            self._add_bp_to_gui(new_id)

    def _update_bp_in_gui(self, i_id: int, i_row: int):
        phrase = matc.settings.get_breathing_phrase(i_id)
        phrase_text: str = f"{phrase.in_breath}\n{phrase.out_breath}"

        qlwi = QtWidgets.QListWidgetItem(phrase_text)

        self.breathing_phrases_qlw.takeItem(i_row)
        self.breathing_phrases_qlw.insertItem(i_row, qlwi)

        self.breathing_phrases_qlw.setCurrentRow(i_row)

    def _add_bp_to_gui(self, i_id: int, i_row: int = NEW_ROW):
        if i_id == matc.globa.BREATHING_PHRASE_NOT_SET:
            phrase_text: str = "nothing"
            raise Exception("BREATHING_PHRASE_NOT_SET --- this should not be possible")
        else:
            phrase = matc.settings.get_breathing_phrase(i_id)
            phrase_text: str = f"{phrase.in_breath}\n{phrase.out_breath}"
        qlwi = QtWidgets.QListWidgetItem(phrase_text)
        if i_id == matc.globa.BREATHING_PHRASE_NOT_SET:
            new_font = qlwi.font()
            new_font.setItalic(True)
            qlwi.setFont(new_font)
        qlwi.setData(QtCore.Qt.UserRole, i_id)
        if i_row == NEW_ROW:
            self.breathing_phrases_qlw.addItem(qlwi)
            new_row: int = self.breathing_phrases_qlw.count() - 1
            self.breathing_phrases_qlw.setCurrentRow(new_row)
        else:
            self.breathing_phrases_qlw.insertItem(i_row, qlwi)
            self.breathing_phrases_qlw.setCurrentRow(i_row)

    def on_del_bp_clicked(self):
        if self.breathing_phrases_qlw.count() == 0:
            QtWidgets.QMessageBox.information(self, "Cannot remove last item",
               "It's not possible to remove the last item")
            return

        current_item = self.breathing_phrases_qlw.currentItem()
        current_row: int = self.breathing_phrases_qlw.currentRow()
        id_: int = current_item.data(QtCore.Qt.UserRole)

        if matc.globa.active_phrase_id == id_:
            QtWidgets.QMessageBox.information(self, "Cannot remove active item",
               "It's not possible to remove the active item. Please switch to another item before removing this one")
            return

        standard_button = QtWidgets.QMessageBox.question(self, "Removing br phrase",
            f"Are you sure you want to remove this item:\n\n{current_item.text()}")
        if standard_button == QtWidgets.QMessageBox.Yes:
            matc.settings.remove_breathing_phrase(id_)
            self.breathing_phrases_qlw.takeItem(current_row)

    def on_edit_bp_clicked(self):
        # bp_edit_dlg = BreathingPhraseEditDialog()
        # bp_edit_dlg.start()
        current_item = self.breathing_phrases_qlw.currentItem()
        current_row: int = self.breathing_phrases_qlw.currentRow()
        id_: int = current_item.data(QtCore.Qt.UserRole)

        result = BreathingPhraseEditDialog.start(id_)
        if result:
            self._update_bp_in_gui(id_, current_row)

        # qlwi = self.breathing_phrases_qlw.takeItem(current_row)
        # self._add_bp_to_gui(id_, current_row)
        # matc.globa.active_phrase_id_it

    def on_br_time_value_changed(self, i_new_value: int):
        if self.updating_gui:
            return
        matc.settings.settings[matc.settings.SK_BREATHING_BREAK_TIMER_SECS] = 60 * i_new_value
        self.br_timer_change_signal.emit()

    def update_menu(self):
        self.menu_bar.clear()

        file_menu = self.menu_bar.addMenu(self.tr("&File"))
        minimize_to_tray_action = QtWidgets.QAction(self.tr("Minimize to tray"), self)
        file_menu.addAction(minimize_to_tray_action)
        minimize_to_tray_action.triggered.connect(self.minimize_to_tray)
        save_action = QtWidgets.QAction(self.tr("Save"), self)
        file_menu.addAction(save_action)
        save_action.triggered.connect(matc.settings.save_settings_to_json_file)
        """
        choose_file_directory_action = QtWidgets.QAction(self.tr("Choose file directory"), self)
        file_menu.addAction(choose_file_directory_action)
        choose_file_directory_action.triggered.connect(pass)
        """
        quit_action = QtWidgets.QAction("Quit application", self)
        file_menu.addAction(quit_action)
        quit_action.triggered.connect(self.exit_application)

        close_settings_action = QtWidgets.QAction("Close settings", self)
        file_menu.addAction(close_settings_action)
        close_settings_action.triggered.connect(self.close)

        debug_menu = self.menu_bar.addMenu("&Debug")
        update_gui_action = QtWidgets.QAction("Update settings GUI", self)
        debug_menu.addAction(update_gui_action)
        update_gui_action.triggered.connect(self.update_gui)
        self.open_settings_dir_action = QtWidgets.QAction("Open settings dir", self)
        debug_menu.addAction(self.open_settings_dir_action)
        self.open_settings_dir_action.triggered.connect(self.on_open_settings_dir_triggered)

        help_menu = self.menu_bar.addMenu(self.tr("&Help"))
        show_intro_dialog_action = QtWidgets.QAction("Show intro wizard", self)
        help_menu.addAction(show_intro_dialog_action)
        show_intro_dialog_action.triggered.connect(self.show_intro_dialog)
        about_action = QtWidgets.QAction(self.tr("About"), self)
        help_menu.addAction(about_action)
        about_action.triggered.connect(self.show_about_box)
        online_help_action = QtWidgets.QAction(self.tr("Online help"), self)
        help_menu.addAction(online_help_action)
        online_help_action.triggered.connect(self.show_online_help)
        feedback_action = QtWidgets.QAction(self.tr("Give feedback"), self)
        help_menu.addAction(feedback_action)
        feedback_action.triggered.connect(self.show_feedback_dialog)
        sysinfo_action = QtWidgets.QAction(self.tr("System Information"), self)
        help_menu.addAction(sysinfo_action)
        sysinfo_action.triggered.connect(self.show_sysinfo_box)

    def on_open_settings_dir_triggered(self):
        path_: str = os.path.dirname(matc.globa.settings_file_path())
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(path_))

    def show_intro_dialog(self):
        self.show_intro_dialog_signal.emit()

    def show_feedback_dialog(self):
        feedback_dlg = FeedbackDialog()
        feedback_dlg.exec_()

    def on_breathing_notification_breathe_clicked(self):
        self.open_breathing_dialog(i_mute_override=True)

    def debug_clear_breathing_phrase_selection(self):
        self.br_phrase_list_wt.list_widget.clearSelection()

    def show_online_help(self):
        url_str = "https://mindfulness-at-the-computer.gitlab.io/user_guide"
        # noinspection PyCallByClass
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(url_str))
        # Python: webbrowser.get(url_str) --- doesn't work

    def show_sysinfo_box(self):
        self._sysinfo_dlg = SysinfoDialog()
        self._sysinfo_dlg.show()

    def show_about_box(self):
        created_by_str = (
            '<p>Originally created by Tord Dellsén - '
            '<a href="https://sunyatazero.gitlab.io/">Website</a></p>'
        )
        all_contributors_str = (
            '<p><a href="https://gitlab.com/mindfulness-at-the-computer/mindfulness-at-the-computer/graphs/master">'
            'All contributors</a></p>'
        )
        photography_str = (
            '<p>Photography for application icon by Torgny Dellsén - '
            '<a href="http://torgnydellsen.zenfolio.com">torgnydellsen.zenfolio.com</a></p>'
        )
        open_iconic_str = (
            '<p>Other icons from Open Iconic - useiconic.com - MIT license</p>'
        )
        about_html_str = (
            '<html>'
            + created_by_str
            + all_contributors_str
            + photography_str
            + open_iconic_str
            + '<p>Other images (for the rest actions) have been released into the public domain (CC0)</p>'
            + '<p>All audio files used have been released into the public domain (CC0)</p>'
            + '<p>Software License: GPLv3 (license text available in the install directory)</p>'
            + '</html>'
        )

        # noinspection PyCallByClass
        QtWidgets.QMessageBox.about(self, "About Mindfulness at the Computer", about_html_str)

    # noinspection PyPep8Naming
    def closeEvent(self, i_QCloseEvent):
        matc.settings.save_settings_to_json_file()
        i_QCloseEvent.ignore()
        self.minimize_to_tray()

    def minimize_to_tray(self):
        self.showMinimized()
        self.hide()

    def exit_application(self):
        QtWidgets.QApplication.quit()
        # sys.exit()

    def update_gui(self):
        self.updating_gui = True

        br_time_value: int = matc.settings.settings[matc.settings.SK_BREATHING_BREAK_TIMER_SECS]
        self.breathing_break_time_qsb.setValue(br_time_value // 60)

        volume: int = matc.settings.settings[matc.settings.SK_BREATHING_AUDIO_VOLUME]
        self.volume_qsr.setValue(volume)

        self.update_gui_active_bold()

        active_br_vis_id: int = matc.settings.settings[matc.settings.SK_BREATHING_VISUALIZATION]
        for btn in self.br_vis_group_qbg.buttons():
            if self.br_vis_group_qbg.id(btn) == active_br_vis_id:
                btn.click()

        move_mouse_cursor: bool = matc.settings.settings[matc.settings.SK_MOVE_MOUSE_CURSOR]
        self.move_mouse_cursor_qcb.setChecked(move_mouse_cursor)

        self.updating_gui = False


if __name__ == "__main__":
    matc_qapplication = QtWidgets.QApplication(sys.argv)
    win = SettingsWin()
    win.show()
    sys.exit(matc_qapplication.exec_())



import os
import enum
from PyQt5 import QtGui
from PyQt5 import QtCore
import matc.constants

#############################################
# This file contains
# * global application state which is not stored on disk
# * global functions, for example:
#   * relating to file names/paths
#   * relating to font
# * enums
# * potentially other global functions
#############################################

NO_PHRASE_SELECTED_INT = -1

GRID_VERTICAL_SPACING_LINUX = 15
BUTTON_BAR_HORIZONTAL_SPACING_LINUX = 2

JSON_SETTINGS_FILE_NAME = "settings.json"
APPLICATION_ICON_NAME_STR = "icon.png"
README_FILE_STR = "README.md"

USER_FILES_DIR = "user_files"
IMAGES_DIR = "images"
ICONS_DIR = "icons"
OPEN_ICONIC_ICONS_DIR = "open_iconic"
AUDIO_DIR = "audio"
RES_DIR = "res"

BIG_BELL_FILENAME = "big_bell[cc0]_fade_out.wav"
SMALL_BELL_SHORT_FILENAME = "small_bell_short[cc0].wav"
BREATHING_PHRASE_NOT_SET: int = -1

LIGHT_GREEN_COLOR = "#bfef7f"
DARK_GREEN_COLOR = "#7fcc19"
DARKER_GREEN_COLOR = "#548811"
WHITE_COLOR = "#ffffff"
BLACK_COLOR = "#1C1C1C"

active_phrase_id = BREATHING_PHRASE_NOT_SET
testing_bool = False
sys_info_telist = []
is_breathing_reminder_shown: bool = False
settings_file_path = ""


class BreathingVisalization(enum.Enum):
    bar = 0
    circle = 1


class FontSize(enum.Enum):
    # Standard font size is (on almost all systems) 12
    small = 9
    medium = 12
    large = 14
    xlarge = 16
    xxlarge = 24


class PhraseSetup(enum.Enum):
    Long = 0
    Switch = 1
    Short = 2


class NotificationType(enum.Enum):
    Both = 0
    Visual = 1
    Audio = 2


class PhraseSelection(enum.Enum):
    same = 0
    random = 1
    ordered = 2


class MoveDirectionEnum(enum.Enum):
    up = 1
    down = 2


def settings_file_exists() -> bool:
    return os.path.isfile(settings_file_path)


def get_config_path(*args) -> str:
    # application_dir_str = os.path.dirname(os.path.dirname(__file__))
    config_dir = QtCore.QStandardPaths.standardLocations(QtCore.QStandardPaths.ConfigLocation)[0]
    # logging.debug("QStandardPaths.ConfigLocation = " + config_dir)

    # There is a bug in Qt:
    # For Windows, the application name is included in QStandardPaths.ConfigLocation
    # For Linux, it's not included
    if matc.constants.APPLICATION_NAME not in config_dir:
        config_dir = os.path.join(config_dir, matc.constants.APPLICATION_NAME)
    full_path_str = config_dir
    for arg in args:
        full_path_str = os.path.join(full_path_str, arg)
    os.makedirs(os.path.dirname(full_path_str), exist_ok=True)
    return full_path_str


def get_settings_file_path(i_date_text: str = ""):
    config_path = get_config_path()
    json_file_name: str = JSON_SETTINGS_FILE_NAME
    if i_date_text:
        json_file_name = json_file_name + "_" + i_date_text
    json_file_path = os.path.join(config_path, json_file_name)
    return json_file_path


def get_module_path() -> str:
    module_dir_str: str = os.path.dirname(os.path.abspath(__file__))
    # base_dir_str: str = os.path.dirname(module_dir_str)
    # base_dir_str = os.getcwd()
    # -__file__ is the file that was started, in other words mindfulness-at-the-computer.py
    return module_dir_str


def get_user_audio_path(i_file_name: str = "") -> str:
    if i_file_name:
        user_audio_path_str = os.path.join(get_module_path(), RES_DIR, AUDIO_DIR, i_file_name)
    else:
        user_audio_path_str = os.path.join(get_module_path(), RES_DIR, AUDIO_DIR)
    return user_audio_path_str


def get_app_icon_path(i_file_name: str) -> str:
    ret_icon_path_str = os.path.join(get_module_path(), RES_DIR, ICONS_DIR, i_file_name)
    return ret_icon_path_str


def get_icon_path(i_file_name: str) -> str:
    ret_icon_path_str = os.path.join(get_module_path(), RES_DIR, ICONS_DIR, "open_iconic", i_file_name)
    return ret_icon_path_str


def get_font(i_size, i_bold: bool = False) -> QtGui.QFont:
    font = QtGui.QFont()
    font.setPointSize(i_size.value)
    font.setBold(i_bold)
    return font


def get_html(i_text: str, i_font_size: int, i_focus: bool = False) -> str:
    ret_str = '<p style="text-align:center;padding:0px;margin:0px;'
    # text_lines_with_br: str = "<br>".join(i_text_lines)
    if i_font_size:
        ret_str += 'font-size:' + str(i_font_size) + 'px;'
    if i_focus:
        ret_str += 'text-decoration: underline;font-weight:bold;'
    ret_str += '">' + i_text + '</p>'
    return ret_str


def clear_widget_and_layout_children(qlayout_or_qwidget) -> None:
    if qlayout_or_qwidget.widget():
        qlayout_or_qwidget.widget().deleteLater()
    elif qlayout_or_qwidget.layout():
        while qlayout_or_qwidget.layout().count():
            child_qlayoutitem = qlayout_or_qwidget.takeAt(0)
            clear_widget_and_layout_children(child_qlayoutitem)  # Recursive call



import logging
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
try:
    # noinspection PyUnresolvedReferences
    from PyQt5 import QtMultimedia
except ImportError:
    logging.debug("ImportError for QtMultimedia - maybe because there's no sound card available")
    # -If the system does not have a sound card (as for example Travis CI)
    # -An alternative to this approach is to use this: http://doc.qt.io/qt-5/qaudiodeviceinfo.html#availableDevices
import matc.constants
import matc.globa
import matc.settings
import matc.gui.breathing_dlg
import matc.gui.settings_win
import matc.gui.intro_dlg


class MainObject(QtCore.QObject):
    """
    We are using this QObject as the core of the application (rather than using a QMainWindow)
    Areas of responsibility:
    * Breathing timer
    * Holds the settings window
    * Holds the breathing-and-rest dialog
    * Handles audio
    * Holds the systray

    """
    def __init__(self):
        super().__init__()
        self.circle_count: int = 1

        # System tray and menu setup
        self.tray_icon = QtWidgets.QSystemTrayIcon(self)
        self.tray_menu = QtWidgets.QMenu()  # self
        self.tray_menu.aboutToShow.connect(self.on_tray_menu_about_to_show)
        self.update_systray_menu()
        self.tray_icon.setContextMenu(self.tray_menu)
        self.tray_icon.messageClicked.connect(self.on_tray_icon_message_clicked)
        self.tray_icon.show()

        # System info (more is available in main.py)
        systray_available_str = "No"
        if self.tray_icon.isSystemTrayAvailable():
            systray_available_str = "Yes"
        matc.globa.sys_info_telist.append(("System tray available", systray_available_str))
        notifications_supported_str = "No"
        if self.tray_icon.supportsMessages():
            notifications_supported_str = "Yes"
        matc.globa.sys_info_telist.append(("System tray notifications supported", notifications_supported_str))
        logging.info("##### System Information #####")
        for (descr_str, value) in matc.globa.sys_info_telist:
            logging.info(descr_str + ": " + str(value))
        logging.info("#####")

        # Audio setup
        self.sound_effect = None
        try:
            self.sound_effect = QtMultimedia.QSoundEffect(self)
            # -PLEASE NOTE: A parent has to be given here, otherwise we will not hear anything
        except NameError:
            logging.debug("NameError - Cannot play audio since QtMultimedia has not been imported")

        # Timer setup
        self.breathing_reminder_timer = Timer(i_continuous=True)
        self.breathing_reminder_timer.timeout_signal.connect(self.on_breathing_timer_timeout)
        """
        self.systray_breathing_timer = Timer(i_continuous=True)
        self.systray_breathing_timer.timeout_signal.connect(self.on_systray_breathing_timeout)
        self.systray_breathing_timer.start(0.32)
        """

        # Window setup
        self.settings_win = None
        self.br_dlg = matc.gui.breathing_dlg.BreathingDlg()
        self.br_dlg.breathing_gv.first_breathing_gi_signal.connect(self.on_first_breathing_gi)
        self.br_dlg.close_signal.connect(self.on_br_dlg_closed)

        # Initialization
        if not matc.globa.settings_file_exists():
            if not matc.globa.testing_bool:
                self.show_intro_dialog()
        self.br_dlg.show_breathing_only()
        self.update_systray_image()

    """

    def on_systray_activated(self, i_reason):
        # LXDE:
        # XFCE:
        # MacOS:
        logging.debug("===== on_systray_activated entered =====")
        logging.debug("i_reason = " + str(i_reason))
        logging.debug("mouseButtons() = " + str(QtWidgets.QApplication.mouseButtons()))
        self.tray_icon.activated.emit(i_reason)

        if i_reason == QtWidgets.QSystemTrayIcon.Trigger:
            self.restore_window()
        else:
            self.tray_icon.activated.emit(i_reason)

        logging.debug("===== on_systray_activated exited =====")
        
    """

    def show_intro_dialog(self):
        self.intro_dlg = matc.gui.intro_dlg.IntroDlg()
        self.intro_dlg.close_signal.connect(self.on_intro_dialog_closed)
        self.intro_dlg.exec()

    def on_intro_dialog_closed(self, i_open_breathing_dialog: bool):
        self._open_br_dlg()

    def _open_br_dlg(self):
        self.br_dlg.show_breathing_only()
        matc.globa.is_breathing_reminder_shown = False
        self.update_systray_image()
        self.breathing_reminder_timer.stop()

    def on_tray_br_phrase_triggered(self, i_action: QtWidgets.QAction):
        if i_action.data():
            matc.globa.active_phrase_id = int(i_action.data())
            self._open_br_dlg()

    def on_tray_quit_triggered(self):
        QtWidgets.QApplication.quit()

    def on_br_dlg_closed(self):
        br_timer_secs: int = matc.settings.settings.get(matc.settings.SK_BREATHING_BREAK_TIMER_SECS)
        self.breathing_reminder_timer.start(br_timer_secs)
        # self.breathing_reminder_timer.start()  # -restarting
        matc.globa.is_breathing_reminder_shown = False
        self.update_systray_image()

    def play_audio(self, i_audio_filename: str, i_volume: int) -> None:
        """
        Please note that the important variable sound_effect is setup at the beginning of the init for settings_win.py
        The reason is that to create this variable we need to send a "self" variable (of which type?)
        :param i_audio_filename:
        :param i_volume:
        :return:
        """
        if self.sound_effect is None:
            logging.warning("play_audio: sound_effect is None")
            return
        audio_path_str = matc.globa.get_user_audio_path(i_audio_filename)
        audio_source_qurl = QtCore.QUrl.fromLocalFile(audio_path_str)
        self.sound_effect.setSource(audio_source_qurl)
        self.sound_effect.setVolume(float(i_volume / 100))
        self.sound_effect.play()

    def on_tray_icon_message_clicked(self):
        print("on_tray_icon_message_clicked")
        """
        Doesn't work on XUbuntu, not sure about other systems
        https://forum.qt.io/topic/115121/qsystemtrayicon-not-sending-messageclicked-signal-on-linux/6
        """
        self._open_br_dlg()

    def on_breathing_timer_timeout(self):
        if self.br_dlg.isVisible():
            return
        if not self.tray_notifications_enabled_action.isChecked():
            return
        # audio_path_str = matc.globa.get_user_audio_path("small_bell_short[cc0].wav")
        volume: int = matc.settings.get_breathing_volume()
        self.play_audio(matc.globa.SMALL_BELL_SHORT_FILENAME, volume)
        matc.globa.is_breathing_reminder_shown = True
        self.update_systray_image()
        ######### self.prepare_timer.start(3)

        phrase = matc.settings.get_breathing_phrase(matc.globa.active_phrase_id)
        phrase_text: str = f"{phrase.in_breath}\n{phrase.out_breath}"

        self.tray_icon.showMessage(matc.constants.APPLICATION_PRETTY_NAME,
            phrase_text, QtWidgets.QSystemTrayIcon.NoIcon, 5000)
        # -on Windows 10, where the tray icon may normally be hidden, it will be shown
        #  during the time thaat the notification popup (message) is shown

    def on_systray_breathing_timeout(self):
        self.update_systray_image(i_increment=True)

    def on_first_breathing_gi(self):
        # audio_path_str = matc.globa.get_user_audio_path("big_bell[cc0].wav")
        volume: int = matc.settings.get_breathing_volume()
        self.play_audio(matc.globa.BIG_BELL_FILENAME, volume)

    def update_systray_menu(self):
        self.tray_open_breathing_dialog_qaction = QtWidgets.QAction(self.tr("Breathing Dialog"))
        self.tray_menu.addAction(self.tray_open_breathing_dialog_qaction)
        self.tray_open_breathing_dialog_qaction.triggered.connect(self.on_tray_open_breathing_dialog_triggered)
        self.tray_br_phrases_qmenu = QtWidgets.QMenu("Phrases")
        self.tray_br_phrases_qmenu.triggered.connect(self.on_tray_br_phrase_triggered)
        self.tray_menu.addMenu(self.tray_br_phrases_qmenu)
        # self.tray_br_phrases_qmenu.aboutToShow.connect(self.on_phrases_submenu_about_to_show)
        self.tray_open_settings_action = QtWidgets.QAction(self.tr("Settings"))
        self.tray_menu.addAction(self.tray_open_settings_action)
        self.tray_open_settings_action.triggered.connect(self.on_tray_open_settings_triggered)

        self.tray_notifications_enabled_action = QtWidgets.QAction(self.tr("Notifications enabled"))
        self.tray_notifications_enabled_action.setCheckable(True)
        self.tray_notifications_enabled_action.setChecked(True)
        self.tray_menu.addAction(self.tray_notifications_enabled_action)
        self.tray_notifications_enabled_action.toggled.connect(
            self.on_tray_notifications_enabled_toggled)

        self.tray_quit_action = QtWidgets.QAction(self.tr("Quit"))
        self.tray_menu.addAction(self.tray_quit_action)
        self.tray_quit_action.triggered.connect(self.on_tray_quit_triggered)

        self.tray_menu.setDefaultAction(self.tray_open_breathing_dialog_qaction)

    def update_systray_image(self, i_increment=False):
        """
        if i_increment:
            self.circle_count += 1
            if self.circle_count >= 33:
                self.circle_count = 1
        icon_file_name_str = f"{int(self.circle_count)}.png"
        """

        icon_file_name_str = "icon.png"
        # if self.br_and_rest_dlg is not None and self.br_and_rest_dlg.isVisible():
        if matc.globa.is_breathing_reminder_shown:
            icon_file_name_str = "icon-b.png"
        systray_icon_path = matc.globa.get_app_icon_path(icon_file_name_str)
        self.tray_icon.setIcon(QtGui.QIcon(systray_icon_path))

    def on_tray_menu_about_to_show(self):
        logging.debug("on_tray_menu_about_to_show")
        # self.rest_progress_qaction.setText("TBD - time since last rest")
        self.tray_br_phrases_qmenu.clear()
        phrases: list[matc.settings.BreathingPhrase] = matc.settings.settings[
            matc.settings.SK_BREATHING_PHRASES]
        for p in phrases:
            if p.id == matc.globa.BREATHING_PHRASE_NOT_SET:
                logging.error("Breathing phrase not set. This should not be possible")
                phrase_text: str = "nothing"
            else:
                phrase = matc.settings.get_breathing_phrase(p.id)
                phrase_text: str = f"{phrase.in_breath}"
            # qlwi = QtWidgets.QListWidgetItem(phrase_text)
            bp_qaction = QtWidgets.QAction(phrase_text, parent=self.tray_br_phrases_qmenu)
            # -please note that we have to use the parent param, otherwise the sub-menu will be empty
            bp_qaction.setData(p.id)
            self.tray_br_phrases_qmenu.addAction(bp_qaction)

    def on_tray_open_settings_triggered(self):
        self.settings_win = matc.gui.settings_win.SettingsWin()
        self.settings_win.show_intro_dialog_signal.connect(self.on_settings_show_intro_dialog)
        self.settings_win.br_timer_change_signal.connect(self.on_settings_br_timer_change)
        self.settings_win.show()

    def on_tray_notifications_enabled_toggled(self, i_checked: bool):
        if i_checked:
            self.breathing_reminder_timer.start()
        else:
            matc.globa.is_breathing_reminder_shown = False
            self.update_systray_image()

    def on_settings_show_intro_dialog(self):
        self.show_intro_dialog()

    def on_settings_br_timer_change(self):
        br_noftif_time = matc.settings.settings[matc.settings.SK_BREATHING_BREAK_TIMER_SECS]
        self.breathing_reminder_timer.start(br_noftif_time)

    def on_tray_open_breathing_dialog_triggered(self):
        self._open_br_dlg()


class Timer(QtCore.QObject):
    timeout_signal = QtCore.pyqtSignal()

    def __init__(self, i_continuous: bool = False):
        super().__init__()
        # self.minutes_elapsed: int = 0
        self.timeout_secs = 0
        self.timer = None
        self.continuous: bool = i_continuous

    def stop(self):
        if self.timer is not None and self.timer.isActive():
            self.timer.stop()
        # self.minutes_elapsed = 0

    def start(self, i_timeout_secs: int = 0):
        self.stop()
        if i_timeout_secs:
            self.timeout_secs = i_timeout_secs
        elif self.timeout_secs:
            pass
        else:
            raise Exception("timeout_secs error, not possible to have 0 for this value")
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.timeout)
        self.timer.start(int(self.timeout_secs * 1000))

    def timeout(self):
        logging.debug("timeout")
        # self.minutes_elapsed += 1
        self.timeout_signal.emit()
        if not self.continuous:
            self.timer.stop()
